"""Contains all reverse-engineering functionality."""


import copy
import warnings
from collections.abc import Container
from tqdm import tqdm
import numpy as np
import scipy.optimize as spo
import scipy.sparse as sps
import iDEA.system
import iDEA.state
import iDEA.observables

def fermi_occs(Eis,mu,tau):
    '''
    INPUT:
        tau: Scalar the tau value to find the temprature dependent mu for
        mus: scalar a guess at a value below the chemical potential for tau
        Eis: Scalar Some kind of energy 
    OUTPUT
        F_occs: Scalar The Fermi Occupation for mu, tau, Eis
    '''
    F_occs = 1/(1+np.exp((Eis-mu)/tau))
    return F_occs

#### Function to calculate number of particles for a given mu and tau (unshifted)
def fermi_particle_number_function(mu,tau,vals,vecs,x):
    '''
    Description: Determine the Unshifted particle number
    INPUT:
        mu: Scalar (float) chemical potential
        tau: Scalar (float) Electronic temperature
        vecs: ndarray Eigenvectors
        vals: ndarray Eigenvalues 
        x: ndarray Grid
    OUTPUT
        Ne: Scalar (float) The unshifted particle number
    '''
    fs = fermi_occs(vals,mu,tau)
    dens = np.zeros(len(x))
    for i,f in enumerate(fs):
        dens += 2*f*vecs[:,i]**2
    Ne = np.trapz(dens,x)
    return Ne

### Function to calculate number of particles for a given mu and tau (shifted)
def fermi_particle_number_shifter(tau,vals,vecs,x,target_Ne): 
    
    '''
    Old title: particle_number_function_function

    Description: Determines the Shifted particle number
    INPUT:
        tau: Scalar (float) Electronic temperature
        vecs: ndarray Eigenvectors
        vals: ndarray Eigenvalues 
        x: ndarray Grid
        target_Ne: Scalar (int) The desired particle number
    OUTPUT
        particle_number_Shift: Scalar (float) The Shifted particle number
    '''

    def fermi_particle_number_Shift(mu):
        
        fs = fermi_occs(vals,mu,tau)
        dens = np.zeros(len(x))
        for i,f in enumerate(fs):
            dens += 2*f*vecs[:,i]**2
        Ne = np.trapz(dens,x)
        
        return Ne - target_Ne

    return fermi_particle_number_Shift

def reverse(
    s: iDEA.system.System,
    target_n: np.ndarray,
    method: Container,
    v_guess: np.ndarray = None,
    mu: float = 1.0,
    pe: float = 0.1,
    tol: float = 1e-12,
    silent: bool = False,
    scheme: int = 1,
    occs_calc: bool = False,
    # Eis: float = 0,
    chem_u: float = 0, 
    tau: float = 0,
    **kwargs
) -> iDEA.state.State:
    r"""
    Determines what ficticious system is needed for a given method, when solving the system, to produce a given target density.
    If the given target density is from solving the interacting electron problem (iDEA.methods.interacting), and the method is the non-interacting electron solver (iDEA.methods.non_interacting)
    the output is the Kohn-Sham system.

    The iterative method used is defined by the following formula:
    .. math:: \mathrm{V}_\mathrm{ext} \rightarrow \mu * (\mathrm{n}^p - \mathrm{target_n}^p)

    | Args:
    |     s: iDEA.system.System, System object.
    |     target_n: np.ndarray, Target density to reverse engineer.
    |     method: Container, The method used to solve the system.
    |     v_guess: np.ndarray, The initial guess of the fictitious potential. (default = None)
    |     mu: float = 1.0, Reverse engineering parameter mu. (default = 1.0)
    |     pe: float = 0.1, Reverse engineering parameter p. (default = 0.1)
    |     tol: float, Tollerance of convergence. (default = 1e-12)
    |     silent: bool, Set to true to prevent printing. (default = False)
    |     scheme: int, Set to 1 to use godby scheme set to 2 to use wang-parr scheme. (default = 1)
    |     occs_calc: bool, Set to True to use the thermal Wang-Parr inversion.  (default = False)
    |     tau: float, Electronic temperature for use with thermal Wang-Parr inversion. 
    |     chem_u: float, chemical potential for use with thermal Wang-Parr inversion. 
    |     placeholder: temp, XXXXXXXX for use with thermal Wang-Parr inversion. 
    |     kwargs: Other arguments that will be given to the method's solve function.

    | Returns:
    |     s_fictitious: iDEA.system.System, fictitious system object.
    """
    s_fictitious = copy.deepcopy(s)
    if v_guess is not None:
        s_fictitious.v_ext = v_guess
    n = np.zeros(shape=s.x.shape)
    up_n = np.zeros(shape=s.x.shape)
    down_n = np.zeros(shape=s.x.shape)
    p = np.zeros(shape=s.x.shape * 2)
    up_p = np.zeros(shape=s.x.shape * 2)
    down_p = np.zeros(shape=s.x.shape * 2)
    convergence = 1.0
    if scheme == 1:
        while convergence > tol:
            if silent is False:
                print(
                    r"iDEA.Godby_inversion: convergence = {:.5}, tolerance = {1:.5}".format(
                        convergence, tol
                    ),
                    end="\r",
                )
            state = method.solve(
                s_fictitious,
                initial=(n, up_n, down_n, p, up_p, down_p),
                silent=True,
                **kwargs
            )
            n, up_n, down_n = iDEA.observables.density(
                s_fictitious, state=state, return_spins=True
            )
            p, up_p, down_p = iDEA.observables.density_matrix(
                s_fictitious, state=state, return_spins=True
            )
            s_fictitious.v_ext += mu * (n**pe - target_n**pe)
            convergence = np.sum(abs(n - target_n)) * s.dx
        if silent is False:
            print()
    if scheme == 2: 
        if occs_calc == False:
            while convergence > tol:
                if silent is False:
                    print(
                        r"iDEA.Wang-Parr_inversion: convergence = {:.5}, tolerance = {1:.5}".format(
                            convergence, tol
                        ),
                        end="\r",
                    )
                state = method.solve(
                    s_fictitious,
                    initial=(n, up_n, down_n, p, up_p, down_p),
                    silent=True,
                    **kwargs
                )
                n, up_n, down_n = iDEA.observables.density(
                    s_fictitious, state=state, return_spins=True
                )
                denom = np.zeros(len(n))
                # print(f"Hello")
                # print(int(s_fictitious.count))
                for i in range(int(s_fictitious.count)):
                    denom += (state.up.orbitals[:,i]**2 + state.down.orbitals[:,i]**2)*(1/(state.up.energies[i]+state.down.energies[i]))
                update_term = ((target_n - n)/(denom))*0.001
                s_fictitious.v_ext += update_term
                convergence = np.sum(abs(n - target_n)) * s.dx
            if silent is False:
                print()
        if occs_calc == True:
            while convergence > tol:
                if silent is False:
                    print(
                        r"iDEA.Wang-Parr_inversion.thermal: convergence = {0:.5}, tolerance = {1:.5}".format(
                            convergence, tol
                        ),
                        end="\r",
                    )
                state = method.solve(
                    s_fictitious,
                    initial=(n, up_n, down_n, p, up_p, down_p),
                    silent=True,
                    **kwargs
                )
                n, up_n, down_n = iDEA.observables.density(
                    s_fictitious, state=state, return_spins=True
                )
                Eis = 2*state.up.energies
                F_occs = 1/(1+np.exp((Eis-chem_u)/tau))
                weighted_density = F_occs * n # need to recalculate weights using the occupation function since ONLY EIGENVALUES will change
                denom = np.zeros(len(n))
                for i in range(int(s_fictitious.count/2)):
                    denom += F_occs * (2*state.up.orbitals[:,i]**2)*(1/(2*state.up.energies[i]))
                update_term = ((target_n - weighted_density)/(denom))*0.001
                s_fictitious.v_ext += update_term
                convergence = np.sum(abs(n - target_n)) * s.dx
            if silent is False:
                print()
    return s_fictitious


def _residual(
    v: np.ndarray,
    s_fictitious: iDEA.system.System,
    evolution_fictitious: iDEA.state.Evolution,
    j: int,
    method: Container,
    v_ptrb: np.ndarray,
    dt: float,
    restricted: bool,
    target_n: np.ndarray,
):
    r"""
    The residual function used to optimise each time step of the time dependent reverse propagation.

    | Args:
    |     v: iDEA.system.System, Potential adjusted during optimisation.
    |     s_fictitious: iDEA.system.System, Fictitious system.
    |     evolution_fictitious: iDEA.system.Evolution, Fictitious evolution.
    |     j: int float = 1.0, Time index.
    |     method: Container: float = 0.1, The method used to solve the system.
    |     v_ptrb: np.ndarray, Local perturbing potential on the grid of t and x values, indexed as v_ptrb[time,space].
    |     dt: float, bool, Timestep.
    |     restricted: bool, Is the calculation restricted (r) on unrestricted (u).
    |     target_n: np.ndarray, Target density.

    | Returns:
    |     residual: np.ndarray, Error in propagation to be minimised.
    """
    v_td = np.zeros_like(v_ptrb)
    v_td[j, :] = v[:]
    evolution = method.propagate_step(
        s_fictitious, evolution_fictitious, j, method.hamiltonian, v_td, dt, restricted
    )
    n = iDEA.observables.density(
        s_fictitious,
        evolution=evolution,
        time_indices=np.array([j]),
        return_spins=False,
    )[0, :]
    residual = n - target_n
    return residual


def reverse_propagation(
    s_fictitious: iDEA.system.System,
    state_fictitious: iDEA.state.State,
    target_n: np.ndarray,
    method: Container,
    v_ptrb: np.ndarray,
    t: np.ndarray,
    restricted: bool = False,
    tol: float = 1e-10,
    **kwargs
) -> iDEA.state.Evolution:
    r"""
    Determines what ficticious evolution is needed for a given method, when solving the system, to produce a given time dependent target density.
    If the given target density is from solving the interacting electron problem (iDEA.methods.interacting), and the method is the non-interacting electron solver (iDEA.methods.non_interacting)
    the output is the Kohn-Sham system.

    | Args:
    |     s_fictitious: iDEA.system.System, System object.
    |     state_fictitious: iDEA.state.State, Fictitious initial state.
    |     target_n: np.ndarray, Target density to reverse engineer.
    |     method: Container, The method used to solve the system.
    |     v_ptrb: np.ndarray, Local perturbing potential on the grid of t and x values, indexed as v_ptrb[time,space].
    |     t: np.ndarray, Grid of time values.
    |     restricted: bool, Is the calculation restricted (r) on unrestricted (u). (default = False)
    |     tol: float, Tollerance of convergence. (default = 1e-10)
    |     kwargs: Other arguments that will be given to the method's solve function.

    | Returns:
    |     evolution_fictitious, error: iDEA.system.Evolution, fictitious evolution object along with time dependent error.
    """
    # Determine the Hamiltonian function.
    hamiltonian_function = method.hamiltonian

    # Construct the unperturbed Hamiltonian.
    n, up_n, down_n = iDEA.observables.density(
        s_fictitious, state=state_fictitious, return_spins=True
    )
    p, up_p, down_p = iDEA.observables.density_matrix(
        s_fictitious, state=state_fictitious, return_spins=True
    )
    H, up_H, down_H = hamiltonian_function(
        s_fictitious, up_n, down_n, up_p, down_p, **kwargs
    )
    H = sps.csc_matrix(H)
    up_H = sps.csc_matrix(up_H)
    down_H = sps.csc_matrix(down_H)
    down_H += sps.spdiags(
        1e-12 * s_fictitious.x,
        np.array([0]),
        s_fictitious.x.shape[0],
        s_fictitious.x.shape[0],
    )

    # Apply restriction.
    if restricted:
        up_H = H
        down_H = H

    # Compute timestep.
    dt = t[1] - t[0]

    # Initilise the single-body time-dependent evolution.
    evolution_fictitious = iDEA.state.SingleBodyEvolution(state_fictitious)
    evolution_fictitious.up.td_orbitals = np.zeros(
        shape=(
            t.shape[0],
            s_fictitious.x.shape[0],
            state_fictitious.up.occupied.shape[0],
        ),
        dtype=complex,
    )
    evolution_fictitious.down.td_orbitals = np.zeros(
        shape=(
            t.shape[0],
            s_fictitious.x.shape[0],
            state_fictitious.down.occupied.shape[0],
        ),
        dtype=complex,
    )
    evolution_fictitious.up.td_orbitals[0, :, :] = state_fictitious.up.orbitals[
        :, state_fictitious.up.occupied
    ]
    evolution_fictitious.down.td_orbitals[0, :, :] = state_fictitious.down.orbitals[
        :, state_fictitious.down.occupied
    ]
    evolution_fictitious.v_ptrb = copy.deepcopy(v_ptrb)
    evolution_fictitious.t = copy.deepcopy(t)

    # Initialise error.
    error = np.zeros_like(t)

    # Reverse propagation.
    for j, ti in enumerate(
        tqdm(
            t,
            desc="iDEA.reverse_engineering.reverse_propagation: reversing propagation",
        )
    ):
        if j != 0:

            # Determine ficticious perturbing potential.
            v_guess = np.zeros_like(evolution_fictitious.v_ptrb[j, :])
            result = spo.root(
                _residual,
                v_guess,
                args=(
                    s_fictitious,
                    evolution_fictitious,
                    j,
                    method,
                    v_ptrb,
                    dt,
                    restricted,
                    target_n[j, :],
                ),
                method="hybr",
                tol=tol,
                options={"maxfev": 200},
            )
            evolution_fictitious.v_ptrb[j, :] = result.x

            # Perform propagation.
            evolution_fictitious = method.propagate_step(
                s_fictitious,
                evolution_fictitious,
                j,
                method.hamiltonian,
                evolution_fictitious.v_ptrb,
                dt,
                restricted,
            )
            n = iDEA.observables.density(
                s_fictitious,
                evolution=evolution_fictitious,
                time_indices=np.array([j]),
                return_spins=False,
            )[0]

            # Compute mae.
            error[j] = np.mean(np.abs(n - target_n[j, :]))

    return evolution_fictitious, error



import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name="iDEA-inversion",
    version="1.0.3",
    author="Jack Wetherell original code, Anthony R. Osborne new inversion code",
    author_email="anthony.r.osborne019@pm.me",
    description="interacting Dynamic Electrons Approach (iDEA) with an additional inversion method packaged only for testing purposes not for distribution",
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    python_requires=">=3.8",
    install_requires=[
        'numpy>=1.22.3',
        'scipy>=1.8.0',
        'matplotlib>=3.5.1',
        'jupyterlab>=3.3.2',
        'tqdm>=4.64.0',
        'black>=22.3.0',
        'autoflake>=1.4.0',
        'build>=0.7.0',
        'twine>=4.0.0',
    ],
)


"""Contains all functionality to define and manage definitions of model systems."""


import warnings
import numpy as np
import iDEA.utilities
import iDEA.interactions


__all__ = ["System", "save_system", "load_system", "systems"]


class System:
    r"""Model system, containing all defining properties."""

    def __init__(
        self,
        x: np.ndarray,
        v_ext: np.ndarray,
        v_int: np.ndarray,
        electrons: str,
        stencil: int = 13,
    ):
        r"""
        Model system, containing all defining properties.

        | Args:
        |     x: np.ndarray, Grid of x values in 1D space.
        |     v_ext: np.ndarray, External potential on the grid of x values.
        |     v_int: np.ndarray, Interaction potential on the grid of x values.
        |     electrons: string, Electrons contained in the system.
        |     stencil: int, Stencil to use for derivatives on the grid of x values. (default = 13)

        | Raises:
        |     AssertionError.
        """
        self.__x = x
        self.__dx = self.x[1] - self.x[0]
        self.v_ext = v_ext
        self.v_int = v_int
        self.__electrons = electrons
        self.count = len(electrons)
        self.up_count = electrons.count("u")
        self.down_count = electrons.count("d")
        self.stencil = stencil
        self.check()

    def check(self):
        r"""Performs checks on system properties. Raises AssertionError if any check fails."""
        assert (
            type(self.x) == np.ndarray
        ), f"x grid is not of type np.ndarray, got {type(self.x)} instead."
        assert (
            type(self.v_ext) == np.ndarray
        ), f"v_ext is not of type np.ndarray, got {type(self.v_ext)} instead."
        assert (
            type(self.v_int) == np.ndarray
        ), f"v_int is not of type np.ndarray, got {type(self.v_int)} instead."
        assert (
            type(self.count) == int
        ), f"count is not of type int, got {type(self.NE)} instead."
        assert (
            len(self.x.shape) == 1
        ), f"x grid is not a 1D array, got {len(self.x.shape)}D array instead."
        assert (
            len(self.v_ext.shape) == 1
        ), f"v_ext is not a 1D array, got {len(self.v_ext.shape)}D array instead."
        assert (
            len(self.v_int.shape) == 2
        ), f"v_int is not a 2D array, got {len(self.v_int.shape)}D array instead."
        assert (
            self.x.shape == self.v_ext.shape
        ), f"x grid and v_ext arrays are not the same shape, got x.shape = {self.x.shape} and v_ext.shape = {self.v_ext.shape} instead."
        assert (
            self.x.shape[0] == self.v_int.shape[0]
            and self.x.shape[0] == self.v_int.shape[1]
        ), "v_int is not of the correct shape, got shape {self.v_int.shape} instead."
        assert self.count >= 0, f"count is not positive."
        assert set(self.electrons).issubset(
            set(["u", "d"])
        ), f"Electrons must have only up or down spin, e.g 'uudd'. Got {self.electrons} instead"
        assert (
            self.count == self.up_count + self.down_count
        ), f"Electrons must obay up_count + down_count = count."
        assert self.stencil in [
            3,
            5,
            7,
            9,
            11,
            13,
        ], f"stencil must be one of [3,5,7,9,11,13], got {self.stencil} instead."

    @property
    def x(self):
        return self.__x

    @x.setter
    def x(self, value):
        self.__x = value
        self.__dx = self.__x[1] - self.__x[0]
        warnings.warn(
            "x grid has been changed: dx has been recomputed, please update v_ext and v_int on this grid."
        )

    @x.deleter
    def x(self):
        del self.__x

    @property
    def dx(self):
        return self.__dx

    @dx.setter
    def dx(self, value):
        raise AttributeError(
            "cannot set dx directly: set the x grid and dx will be updated automatically."
        )

    @dx.deleter
    def dx(self):
        del self.__dx

    @property
    def electrons(self):
        return self.__electrons

    @electrons.setter
    def electrons(self, value):
        self.__electrons = value
        self.count = len(value)
        self.up_count = value.count("u")
        self.down_count = value.count("d")

    @electrons.deleter
    def electrons(self):
        del self.__electrons

    def __str__(self):
        return f"iDEA.system.System: x = np.array([{self.x[0]:.3f},...,{self.x[-1]:.3f}]), dx = {self.dx:.4f}..., v_ext = np.array([{self.v_ext[0]:.3f},...,{self.v_ext[-1]:.3f}]), electrons = {self.electrons}"


def save_system(s: System, file_name: str) -> None:
    r"""
    Save a system to an system file.

    | Args:
    |     system: iDEA.system.System, System object to save.
    |     file_name: str, file name.
    """
    pickle.dump(s, open(file_name, "wb"))


def load_system(file_name: str) -> System:
    r"""
    Load a system from an system file.

    | Args:
    |     file_name: str, file name.

    | Returns
    |     system: iDEA.system.System, Loaded System object.
    """
    return pickle.load(open(file_name, "rb"))


# Define some default built in systems.
__x1 = np.linspace(-10, 10, 300)
systems = iDEA.utilities.Container()
systems.qho = System(
    __x1,
    0.5 * (0.25**2) * (__x1**2),
    iDEA.interactions.softened_interaction(__x1),
    "uu",
)
__x2 = np.linspace(-20, 20, 300)
systems.atom = System(
    __x2, -2.0 / (abs(__x2) + 1.0), iDEA.interactions.softened_interaction(__x2), "ud"
)



"""Contains some pre-defined electron-electron interations"""


import numpy as np


__all__ = [
    "softened_interaction",
    "softened_interaction_alternative",
    "raw_interaction",
]


def softened_interaction(
    x: np.ndarray, strength: float = 1.0, softening: float = 1.0
) -> np.ndarray:
    r"""
    Constructs the softened interaction potential.

    .. math:: v_\mathrm{int}(x,x') = \frac{s}{|x-x'| + a}


    | Args:
    |     x: np.ndarray, x grid.
    |     strength: float, Strength of the interaction .. math:: s. (default = 1.0)
    |     softening: float, Softening parameter of the interaction .. math:: a. (default = 1.0)

    | Returns:
    |     v_int: np.ndarray, Softened interaction potential on x grid of the System.
    """
    v_int = np.zeros((x.shape[0], x.shape[0]), dtype="float")
    for i in range(x.shape[0]):
        for j in range(x.shape[0]):
            v_int[i, j] = strength / (abs(x[i] - x[j]) + softening)
    return v_int


def softened_interaction_alternative(
    x: np.ndarray, strength: float = 1.0, softening: float = 1.0
) -> np.ndarray:
    r"""
    Constructs the alternative softened interaction potential.

    .. math:: v_\mathrm{int}(x,x') = \frac{s}{{(\sqrt{x-x'} + a)}^{2}}

    | Args:
    |     x: np.ndarray, x grid.
    |     strength: float, Strength of the interaction .. math:: s. (default = 1.0)
    |     softening: float, Softening parameter of the interaction .. math:: a. (default = 1.0)

    | Returns:
    |     v_int: np.ndarray, Softened interaction potential on x grid of the System.
    """
    v_int = np.zeros((x.shape[0], x.shape[0]), dtype="float")
    for i in range(x.shape[0]):
        for j in range(x.shape[0]):
            v_int[i, j] = strength / np.sqrt(((x[i] - x[j]) ** 2 + softening))
    return v_int


def raw_interaction(x: np.ndarray, strength: float = 1.0) -> np.ndarray:
    r"""
    Constructs the raw interaction potential.

    .. math:: v_\mathrm{int}(x,x') = \frac{s}{|x-x'| + 10^{-10}}

    | Args:
    |     x: np.ndarray, x grid.
    |     strength: float, Strength of the interaction .. math:: s. (default = 1.0)

    | Returns:
    |     v_int: np.ndarray, Softened interaction potential on x grid of the System.
    """
    return softened_interaction(x, strength=strength, softening=1e-10)



"""Contains all interacting functionality and solvers."""


import os
import copy
import string
import itertools
import functools
from tqdm import tqdm
import numpy as np
import scipy.sparse as sps
import scipy.sparse.linalg as spsla
import iDEA.system
import iDEA.state
import iDEA.methods.non_interacting


if os.environ.get("iDEA_GPU") == "True":
    import cupy as cnp
    import cupyx.scipy.sparse as csps
    import cupyx.scipy.sparse.linalg as cspsla


name = "interacting"


def kinetic_energy_operator(s: iDEA.system.System) -> sps.dia_matrix:
    r"""
    Compute many-particle kinetic energy operator as a matrix.

    This is built using a given number of finite differences to represent the second derivative.
    The number of differences taken is defined in s.stencil.

    | Args:
    |     s: iDEA.system.System, System object.

    | Returns:
    |     K: sps.dia_matrix, Kintetic energy operator.
    """
    k = iDEA.methods.non_interacting.kinetic_energy_operator(s)
    k = sps.dia_matrix(k)
    I = sps.identity(s.x.shape[0], format="dia")
    partial_operators = lambda A, B, k, n: (
        A if i + k == n - 1 else B for i in range(n)
    )
    fold_partial_operators = lambda f, po: functools.reduce(
        lambda acc, val: f(val, acc, format="dia"), po
    )
    generate_terms = lambda f, A, B, n: (
        fold_partial_operators(f, partial_operators(A, B, k, n)) for k in range(n)
    )
    terms = generate_terms(sps.kron, h, I, s.count)
    K = sps.dia_matrix((s.x.shape[0] ** s.count,) * 2, dtype=float)
    for term in terms:
        K += term
    return K


def external_potential_operator(s: iDEA.system.System) -> sps.dia_matrix:
    r"""
    Compute many-particle external potential energy operator as a matrix.

    | Args:
    |     s: iDEA.system.System, System object.

    | Returns:
    |     Vext: sps.dia_matrix, External potential operator.
    """
    vext = iDEA.methods.non_interacting.external_potential_operator(s)
    vext = sps.dia_matrix(vext)
    I = sps.identity(s.x.shape[0], format="dia")
    partial_operators = lambda A, B, k, n: (
        A if i + k == n - 1 else B for i in range(n)
    )
    fold_partial_operators = lambda f, po: functools.reduce(
        lambda acc, val: f(val, acc, format="dia"), po
    )
    generate_terms = lambda f, A, B, n: (
        fold_partial_operators(f, partial_operators(A, B, k, n)) for k in range(n)
    )
    terms = generate_terms(sps.kron, h, I, s.count)
    Vext = sps.dia_matrix((s.x.shape[0] ** s.count,) * 2, dtype=float)
    for term in terms:
        Vext += term
    return Vext


def hamiltonian(s: iDEA.system.System) -> sps.dia_matrix:
    r"""
    Compute the many-body Hamiltonian.

    | Args:
    |     s: iDEA.system.System, System object.

    | Returns:
    |     H: sps.dia_matrix, Hamiltonian.
    """
    # Construct the non-interacting part of the many-body Hamiltonian
    h = iDEA.methods.non_interacting.hamiltonian(s)[0]
    h = sps.dia_matrix(h)
    I = sps.identity(s.x.shape[0], format="dia")
    partial_operators = lambda A, B, k, n: (
        A if i + k == n - 1 else B for i in range(n)
    )
    fold_partial_operators = lambda f, po: functools.reduce(
        lambda acc, val: f(val, acc, format="dia"), po
    )
    generate_terms = lambda f, A, B, n: (
        fold_partial_operators(f, partial_operators(A, B, k, n)) for k in range(n)
    )
    terms = generate_terms(sps.kron, h, I, s.count)
    H0 = sps.dia_matrix((s.x.shape[0] ** s.count,) * 2, dtype=float)
    for term in terms:
        H0 += term

    # Add the interaction part of the many-body Hamiltonian
    symbols = string.ascii_lowercase + string.ascii_uppercase
    if s.count > 1:
        indices = ",".join(
            ["".join(c) for c in itertools.combinations(symbols[: s.count], 2)]
        )
        U = np.log(
            np.einsum(
                indices + "->" + symbols[: s.count],
                *(np.exp(s.v_int),) * int(s.count * (s.count - 1) / 2)
            )
        )
        U = sps.diags(U.reshape((H0.shape[0])), format="dia")
    else:
        U = 0.0

    # Construct the total many-body Hamiltonian
    H = H0 + U

    return H


def total_energy(s: iDEA.system.System, state: iDEA.state.ManyBodyState) -> float:
    r"""
    Compute the total energy of an interacting state.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.ManyBodyState, State.

    | Returns:
    |     E: float, Total energy.
    """
    return state.energy


def _permutation_parity(p):
    r"""
    Compute the permulation paritiy of a given permutation.

    | Args:
    |     p: tuple, Permutation.

    | Returns:
    |     parity: float, Permutation parity.
    """
    p = list(p)
    parity = 1
    for i in range(0, len(p) - 1):
        if p[i] != i:
            parity *= -1
            mn = min(range(i, len(p)), key=p.__getitem__)
            p[i], p[mn] = p[mn], p[i]
    return parity


def antisymmetrize(s, spaces, spins, energies):
    r"""
    Antisymmetrize the solution to the Schrodinger equation.

    | Args:
    |     s: iDEA.system.System, System object.
    |     spaces: np.ndarray, Spatial parts of the wavefunction.
    |     spins: np.ndarray, Spin parts of the wavefunction.
    |     energies: np.ndarray, Energies.

    | Returns:
    |     fulls: np.ndarray, Full anantisymmetrized wavefunction.
    |     spaces: np.ndarray, Spatial parts of the wavefunction.
    |     spins: np.ndarray, Spin parts of the wavefunction.
    |     energies: np.ndarray, Energies.

    """
    # Perform antisymmetrization.
    l = string.ascii_lowercase[: s.count]
    L = string.ascii_uppercase[: s.count]
    st = (
        l
        + "Y,"
        + L
        + "Y->"
        + "".join([i for sub in list(zip(l, L)) for i in sub])
        + "Y"
    )
    fulls = np.einsum(st, spaces, spins)
    # print(fulls[...,-1])
    L = list(zip(list(range(0, s.count * 2, 2)), list(range(1, s.count * 2, 2))))
    perms = itertools.permutations(list(range(s.count)))
    fulls_copy = copy.deepcopy(fulls)
    fulls = np.zeros_like(fulls)
    for p in perms:
        indices = list(itertools.chain(*[L[e] for e in p]))
        fulls += _permutation_parity(p) * np.moveaxis(
            fulls_copy, list(range(s.count * 2)), indices
        )

    # Filter out zeros.
    allowed_fulls = []
    allowed_energies = []
    allowed_spaces = []
    allowed_spins = []
    for n in range(fulls.shape[-1]):
        if np.allclose(fulls[..., n], np.zeros(fulls.shape[:-1])):
            pass
        else:
            allowed_fulls.append(fulls[..., n])
            allowed_energies.append(energies[n])
            allowed_spaces.append(spaces[..., n])
            allowed_spins.append(spins[..., n])
    fulls = np.moveaxis(np.array(allowed_fulls), 0, -1)
    # print(fulls)
    spaces = np.moveaxis(np.array(allowed_spaces), 0, -1)
    spins = np.moveaxis(np.array(allowed_spins), 0, -1)
    energies = np.array(allowed_energies)

    # Normalise.
    # print(fulls.shape)
    for k in range(fulls.shape[-1]):
        # print(np.max(fulls[..., k]))
        fulls[..., k] = fulls[..., k] / np.sqrt(
            np.sum(abs(fulls[..., k]) ** 2) * s.dx**s.count
        )

    # Filter out duplicates.
    allowed_fulls = []
    allowed_energies = []
    # print("HELLOOO????")
    # print(fulls.shape )
    for n in range(fulls.shape[-1] - 1):
        if np.allclose(abs(fulls[..., n]), abs(fulls[..., n + 1])):
            # print(fulls)
            pass
        else:
            # print(fulls)
            allowed_fulls.append(fulls[..., n])
            allowed_energies.append(energies[n])
    # print(fulls)
    allowed_fulls.append(fulls[..., -1])
    allowed_energies.append(energies[-1])
    fulls = np.moveaxis(np.array(allowed_fulls), 0, -1)
    spaces = spaces[..., : fulls.shape[-1]]
    spins = spins[..., : fulls.shape[-1]]
    energies = np.array(allowed_energies)

    return fulls, spaces, spins, energies


def _estimate_level(s: iDEA.system.System, k: int) -> int:
    r"""
    Estimate the solution to the Schrodinger equation needed to eachive given antisymetric energy state.

    | Args:
    |     s: iDEA.system.System, System object.
    |     k: int, Target energy state.

    | Returns:
    |     level: int, Extimate of level of excitement.
    """
    return (abs(s.up_count - s.down_count) + 1) ** 2 * s.count * (k + 1)


def _solve_on_gpu(H: np.ndarray, k: int) -> tuple:
    r"""
    Solves the eigenproblem on the GPU.

    | Args:
    |     H: np.ndarray, Hamiltonian.
    |     k: int, Eigenstate to solve for.

    | Returns:
    |     eigenvalues_gpu, eigenstates_gpu: tuple, Solved eigenvalues and eigenstates.
    """
    sigma = 0
    which = "LA"
    H_gpu_shifted = csps.csr_matrix(H - sigma * csps.csr_matrix(sps.eye(H.shape[0])))
    H_gpu_LU = cspsla.splu(H_gpu_shifted)
    H_gpu_LO = cspsla.LinearOperator(H_gpu_shifted.shape, H_gpu_LU.solve)
    eigenvalues_gpu, eigenstates_gpu = cspsla.eigsh(H_gpu_LO, k=k, which=which)
    eigenvalues_gpu = eigenvalues_gpu
    eigenstates_gpu = eigenstates_gpu
    eigenvalues_gpu = (1 + eigenvalues_gpu * sigma) / eigenvalues_gpu
    idx = np.argsort(eigenvalues_gpu)
    eigenstates_gpu = cnp.transpose(eigenstates_gpu)
    eigenvalues_gpu = eigenvalues_gpu[idx]
    eigenstates_gpu = cnp.transpose(eigenstates_gpu[idx])
    return eigenvalues_gpu, eigenstates_gpu


def solve(
    s: iDEA.system.System, H: np.ndarray = None, k: int = 0, level=None
) -> iDEA.state.ManyBodyState:
    r"""
    Solves the interacting Schrodinger equation of the given system.

    | Args:
    |     s: iDEA.system.System, System object.
    |     H: np.ndarray, Hamiltonian [If None this will be computed from s]. (default = None)
    |     k: int, Energy state to solve for. (default = 0, the ground-state)
    |     level: int. Max level of excitation to use when solving the Schrodinger equation.

    | Returns:
    |     state: iDEA.state.ManyBodyState, Solved state.
    """
    # Construct the many-body state.
    state = iDEA.state.ManyBodyState()

    # Construct the Hamiltonian.
    if H is None:
        H = hamiltonian(s)

    # Estimate the level of excitation.
    if level is None:
        level = _estimate_level(s, k)

    # Solve the many-body Schrodinger equation.
    print("iDEA.methods.interacting.solve: solving eigenproblem...")
    if os.environ.get("iDEA_GPU") == "True":
        H_gpu = csps.csr_matrix(H)
        energies, spaces = _solve_on_gpu(H_gpu, level)
        energies = energies.get()
        spaces = spaces.get()
    else:
        energies, spaces = spsla.eigsh(H.tocsr(), k=level, which="SA")

    # Reshape and normalise the solutions.
    spaces = spaces.reshape((s.x.shape[0],) * s.count + (spaces.shape[-1],))
    for j in range(spaces.shape[-1]):
        spaces[..., j] = spaces[..., j] / np.sqrt(
            np.sum(abs(spaces[..., j]) ** 2) * s.dx**s.count
        )

    # Construct the spin part.
    symbols = string.ascii_lowercase + string.ascii_uppercase
    u = np.array([1, 0])
    d = np.array([0, 1])
    spin_state = tuple([u if spin == "u" else d for spin in s.electrons])
    spin = np.einsum(
        ",".join(symbols[: s.count]) + "->" + "".join(symbols[: s.count]), *spin_state
    )
    spins = np.zeros(shape=((2,) * s.count + (spaces.shape[-1],)))
    for i in range(spaces.shape[-1]):
        spins[..., i] = spin

    # Antisymmetrize.
    fulls, spaces, spins, energies = antisymmetrize(s, spaces, spins, energies)

    # Populate the state.
    state.space = spaces[..., k]
    state.spin = spins[..., k]
    state.full = fulls[..., k]
    state.energy = energies[k]

    return state


def propagate_step(
    s: iDEA.system.System,
    evolution: iDEA.state.ManyBodyEvolution,
    H: sps.dia_matrix,
    v_ptrb: np.ndarray,
    j: int,
    dt: float,
    objs: tuple,
) -> iDEA.state.ManyBodyEvolution:
    r"""
    Propagate a many body state forward in time, one time-step, due to a local pertubation.

    | Args:
    |     s: iDEA.system.System, System object.
    |     evolution: iDEA.state.ManyBodyEvolution, time-dependent evolution.
    |     H: np.ndarray, Static Hamiltonian [If None this will be computed from s]. (default = None)
    |     v_ptrb: np.ndarray, Local perturbing potential on the grid of t and x values, indexed as v_ptrb[time,space].
    |     j: int, Time index.
    |     dt: float, Time-step.
    |     objs: tuple. Tuple of objects needed to construct many-body operator (I, generate_terms).

    | Returns:
    |     evolution: iDEA.state.ManyBodyEvolution, time-dependent evolution one time-step evolved.
    """
    # Construct the pertubation potential.
    vptrb = sps.dia_matrix(np.diag(v_ptrb[j, :]))
    terms = objs[1](sps.kron, vptrb, objs[0], s.count)
    Vptrb = sps.dia_matrix((s.x.shape[0] ** s.count,) * 2, dtype=float)
    for term in terms:
        Vptrb += term

    # Contruct the perturbed Hamiltonian.
    Hp = H + Vptrb

    # Evolve.
    wavefunction = evolution.td_space[j - 1, ...].reshape((s.x.shape[0] ** s.count))
    wavefunction = spsla.expm_multiply(-1.0j * dt * Hp, wavefunction)
    evolution.td_space[j, ...] = wavefunction.reshape((s.x.shape[0],) * s.count)

    return evolution


def propagate(
    s: iDEA.system.System,
    state: iDEA.state.ManyBodyState,
    v_ptrb: np.ndarray,
    t: np.ndarray,
    H: sps.dia_matrix = None,
) -> iDEA.state.ManyBodyEvolution:
    r"""
    Propagate a many body state forward in time due to a local pertubation.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.ManyBodyState, State to be propigated.
    |     v_ptrb: np.ndarray, Local perturbing potential on the grid of t and x values, indexed as v_ptrb[time,space].
    |     t: np.ndarray, Grid of time values.
    |     H: np.ndarray, Static Hamiltonian [If None this will be computed from s]. (default = None)

    | Returns:
    |     evolution: iDEA.state.ManyBodyEvolution, Solved time-dependent evolution.
    """
    # Construct the unperturbed Hamiltonian.
    if H is None:
        H = hamiltonian(s)

    # Compute timestep.
    dt = t[1] - t[0]

    # Initilise time-dependent wavefunction.
    evolution = iDEA.state.ManyBodyEvolution(initial_state=state)
    evolution.td_space = np.zeros(shape=t.shape + state.space.shape, dtype=complex)
    evolution.td_space[0, ...] = copy.deepcopy(evolution.space)

    # Construct objects needed to update potential.
    I = sps.identity(s.x.shape[0], format="dia")
    partial_operators = lambda A, B, k, n: (
        A if i + k == n - 1 else B for i in range(n)
    )
    fold_partial_operators = lambda f, po: functools.reduce(
        lambda acc, val: f(val, acc, format="dia"), po
    )
    generate_terms = lambda f, A, B, n: (
        fold_partial_operators(f, partial_operators(A, B, k, n)) for k in range(n)
    )
    objs = (I, generate_terms)

    # Propagate.
    for j, ti in enumerate(
        tqdm(t, desc="iDEA.methods.interacting.propagate: propagating state")
    ):
        if j != 0:
            propagate_step(s, evolution, H, v_ptrb, j, dt, objs)

    # Populate the many-body time-dependent evolution.
    evolution.v_ptrb = v_ptrb
    evolution.t = t

    return evolution
